Skip to contentSkip to conversation

net: skb_copy_datagram_from_iovec()

PATCH Applied Rusty Russell 4 participants 7 messages Aug 12, 2008 – Aug 15, 2008

Subsystems (3)
lore
Share & tools
lore
#0 lore raw

PATCH 1/2 net: skb_copy_datagram_from_iovec()

There's an skb_copy_datagram_iovec() to copy out of a paged skb, but nothing the other way around (because we don't do that).

We want to allocate big skbs in tun.c, so let's add the function. It's a carbon copy of skb_copy_datagram_iovec() with enough changes to be annoying.

---
 include/linux/skbuff.h |    4 ++
 net/core/datagram.c    |   87 +++++++++++++++++++++++++++++++++++++++++++++++++
 2 files changed, 91 insertions(+)
include/linux/skbuff.h +4 -0
diff -r fcbf6d08c910 include/linux/skbuff.h
@@ -1452,6 +1452,10 @@ extern int	       skb_copy_and_csum_data
 extern int	       skb_copy_and_csum_datagram_iovec(struct sk_buff *skb,
 							int hlen,
 							struct iovec *iov);
+extern int	       skb_copy_datagram_from_iovec(struct sk_buff *skb,
+						    int offset,
+						    struct iovec *from,
+						    int len);
 extern void	       skb_free_datagram(struct sock *sk, struct sk_buff *skb);
 extern int	       skb_kill_datagram(struct sock *sk, struct sk_buff *skb,
 					 unsigned int flags);
net/core/datagram.c +87 -0

93 lines not on this page · raw message

#1 lore raw

PATCH 2/2 tun: fallback if skb_alloc() fails on big packets

skb_alloc produces linear packets (using kmalloc()). That can fail, so should we fall back to making paged skbs.

My original version of this patch always allocate paged skbs for big packets. But that made performance drop from 8.4 seconds to 8.8 seconds on 1G lguest->Host TCP xmit. So now we only do that as a fallback.

drivers/net/tun.c +62 -4
diff -r ffcd4a3f63a8 drivers/net/tun.c
@@ -358,6 +358,66 @@ static unsigned int tun_chr_poll(struct 
 	return mask;
 }
 
+/* prepad is the amount to reserve at front.  len is length after that.
+ * linear is a hint as to how much to copy (usually headers). */
+static struct sk_buff *tun_alloc_skb(size_t prepad, size_t len, size_t linear,
+				     gfp_t gfp)
+{
+	struct sk_buff *skb;
+	unsigned int i;
+
+	skb = alloc_skb(prepad + len, gfp|__GFP_NOWARN);
+	if (skb) {
+		skb_reserve(skb, prepad);
+		skb_put(skb, len); 
+		return skb;
+	}
+
+	/* Under a page?  Don't bother with paged skb. */
+	if (prepad + len < PAGE_SIZE)
+		return NULL;
+
+	/* Start with a normal skb, and add pages. */
+	skb = alloc_skb(prepad + linear, gfp);
+	if (!skb)
+		return NULL;
+
+	skb_reserve(skb, prepad);
+	skb_put(skb, linear); 
+
+	len -= linear;
+
+	for (i = 0; i < MAX_SKB_FRAGS; i++) {
+		skb_frag_t *f = &skb_shinfo(skb)->frags[i];
+
+		f->page = alloc_page(gfp|__GFP_ZERO);
+		if (!f->page)
+			break;
+
+		f->page_offset = 0;
+		f->size = PAGE_SIZE;
+
+		skb->data_len += PAGE_SIZE;
+		skb->len += PAGE_SIZE;
+		skb->truesize += PAGE_SIZE;
+		skb_shinfo(skb)->nr_frags++;
+
+		if (len < PAGE_SIZE) {
+			len = 0;
+			break;
+		}
+		len -= PAGE_SIZE;
+	}
+
+	/* Too large, or alloc fail? */
+	if (unlikely(len)) {
+		kfree_skb(skb);
+		skb = NULL;
+	}
+
+	return skb;
+}
+
 /* Get packet from user space buffer */
 static __inline__ ssize_t tun_get_user(struct tun_struct *tun, struct iovec *iv, size_t count)
 {
@@ -391,14 +451,12 @@ static __inline__ ssize_t tun_get_user(s
 			return -EINVAL;
 	}
 
-	if (!(skb = alloc_skb(len + align, GFP_KERNEL))) {
+	if (!(skb = tun_alloc_skb(align, len, gso.hdr_len, GFP_KERNEL))) {
 		tun->dev->stats.rx_dropped++;
 		return -ENOMEM;
 	}
 
-	if (align)
-		skb_reserve(skb, align);
-	if (memcpy_fromiovec(skb_put(skb, len), iv, len)) {
+	if (skb_copy_datagram_from_iovec(skb, 0, iv, len)) {
 		tun->dev->stats.rx_dropped++;
 		kfree_skb(skb);
 		return -EFAULT;
#2 lore raw

Re: [PATCH 2/2] tun: fallback if skb_alloc() fails on big packets

Quoted 9 lines · Rusty Russell

skb_alloc produces linear packets (using kmalloc()). That can fail, so should we fall back to making paged skbs.

My original version of this patch always allocate paged skbs for big packets. But that made performance drop from 8.4 seconds to 8.8 seconds on 1G lguest->Host TCP xmit. So now we only do that as a fallback.

I'm not sure that this is really a good idea. If anything then tries to expand the head of this skb, they may fail and be forced to drop the packet.

Cheers,

Signature · 4 lines
Visit Openswan at http://www.openswan.org/
Email: Herbert Xu ~{PmV>HI~} <[email protected]>
Home Page: http://gondor.apana.org.au/~herbert/
PGP Key: http://gondor.apana.org.au/~herbert/pubkey.txt
#3 lore raw

Re: [PATCH 2/2] tun: fallback if skb_alloc() fails on big packets

Quoted 7 lines · On Tuesday 12 August 2008 20:14:09 Herbert Xu wrote:
Quoted 2 lines · Rusty Russell

skb_alloc produces linear packets (using kmalloc()). That can fail, so should we fall back to making paged skbs.

I'm not sure that this is really a good idea. If anything then tries to expand the head of this skb, they may fail and be forced to drop the packet.

Yes, but it's no worse than now. virtio_net keeps a cache of allocated pages, but that's more code; and if I'm going to generalize that I really should create a shrinker callback, which produces locking issues.

So I decided this was probably enough for this merge window. Rusty.

#4 lore raw

Re: [PATCH 2/2] tun: fallback if skb_alloc() fails on big packets

Quoted 11 lines · Rusty Russell
Quoted 6 lines · On Tuesday 12 August 2008 20:14:09 Herbert Xu wrote:
Quoted 2 lines · Rusty Russell

skb_alloc produces linear packets (using kmalloc()). That can fail, so should we fall back to making paged skbs.

I'm not sure that this is really a good idea. If anything then tries to expand the head of this skb, they may fail and be forced to drop the packet.

Yes, but it's no worse than now. virtio_net keeps a cache of allocated pages, but that's more code; and if I'm going to generalize that I really should create a shrinker callback, which produces locking issues.

I agree with Rusty. It's no worse than what we have now (ie flat by default) and improves the case were large GSO would otherwise cause packet drops.

TUN changes look good to me. Ack.
I did not get a chance to read through the new skb_copy_datagram_from_iovec() code. At first glance looks good.

Max

maintainer Applied
#5 lore raw

Re: [PATCH 2/2] tun: fallback if skb_alloc() fails on big packets

From: Max Krasnyansky <[email protected]>
Date: Fri, 15 Aug 2008 11:09:07 -0700

Quoted 1 line

TUN changes look good to me. Ack.

I've applied Rusty's patch. I have no suggestions on how to handle his situation better :)

maintainer Applied
#6 lore raw
In reply to Rusty Russell

From: Rusty Russell <[email protected]>
Date: Tue, 12 Aug 2008 16:24:56 +1000

Quoted 8 lines

There's an skb_copy_datagram_iovec() to copy out of a paged skb, but nothing the other way around (because we don't do that).

We want to allocate big skbs in tun.c, so let's add the function. It's a carbon copy of skb_copy_datagram_iovec() with enough changes to be annoying.

Applied.